# This source code is part of the Gecos package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__version__ = "2.0.0"
__author__ = "Patrick Kunzmann"

from .colors import *
from .space import *
from .optimizer import *
from .file import *
